//
//  Storyboard.h
//  Storyboard
//
//  Created by Eliav on 24/04/2019.
//  Copyright © 2019 Glassbox. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <UIKit/UIViewController.h>

extern double const GB_SDK_VER;

@protocol GLASessionDelegate <NSObject>

@optional

/**
 Called when configuration is received successfully
 @param configuration The received configuration
 */
- (void)sessionStartedWithConfiguration:(NSDictionary *)configuration;

/**
 Called when failed receiving configuration
 */
- (void)sessionFailedWithError:(NSError *)error;

/**
 Called when configuration is received successfully but the session is not recorded, for example due to configured device filter
 @param reason Why the session is not recorded
 @param configuration The received configuration
 */
- (void)sessionExcludedWithReason:(NSString *)reason withConfiguration:(NSDictionary *)configuration;

@end


/**
 * Controls configuration environment mode.
 * 1. EXTERNAL - Default configuration mode.
 * 2. INTERNAL - use when configuration is hosted on a private server.
 */
typedef NS_ENUM(NSUInteger, GLAConfigurationType) {
    GLAConfigurationExternal = 0,
    GLAConfigurationInternal
};

@interface GLAStartupSettings :NSObject

+ (instancetype)settings;

/**
* server url.
*/
@property NSString *reportUrl;

/**
 * Optional url to fetch sdk configuration from, if not set the report url is used.
 */
@property NSString *appConfigUrl;

/**
*  application id.
*/
@property NSString *appId;

/**
* certificate for certificate pinning.
*/
@property NSData *certificate;

/**
* support both web and native analytics
*/
@property BOOL hybridMode;

/**
* Controls configuration environment mode.
*/
@property GLAConfigurationType configuration;

/**
* Delete any data that was recorded before receiving the first configuration from the server.
*/
@property (nonatomic) BOOL omitMonitoringBeforeConfig;


/**
* Disable the addition of mobile sessionId to report query string.
*/
@property (nonatomic) BOOL disableNewSessionizing;

@end

/**
 * Controls touch position within a sensitive view.
 * 1. CENTER - Replace the touch location with the center of the view
 * 2. NONE - Removes the touch location completely
 * 3. DISPLAY - No changes (default)
 */
typedef NS_ENUM(NSUInteger, GLATouchMode) {
    GLA_CENTER = 0,
    GLA_NONE,
    GLA_DISPLAY
};

@interface GLAVisibilityFlags : NSObject

/**
 *  Default visibility flags
 */
+ (instancetype)flags;

/**
 * Create visibility flags from a dictionary
 */
+ (instancetype)flagsFromDictionary:(NSDictionary *)dictionary;

/**
 * Indicates if UIView elements with alpha property set to 0 should be masked.
 * Default is NO.
 */
@property BOOL maskOnAlpha;

/**
 * Indicates if UIView elements with isHidden property set to true should be masked.
 * Default is NO.
 */
@property BOOL maskOnHidden;

/**
 * Indicates if UIView which reside on a UIViewController which is not top view controller should be masked.
 * Default is YES.
 */
@property BOOL maskOnAllControllers;


/**
 * Hides all views with an equal tag to the provided view.
 */
@property BOOL maskByTag;

/**
 * Omit analytics info for touch event
 */
@property BOOL omitAnalytics;


@property GLATouchMode touchMode;

@end

@interface Storyboard : NSObject

/**
 * @deprecated Please use the start:settings method
 * @function start:reportUrl:appId:userData:error
 * @brief Starts the monitoring with user data.
 * The user data in this case can replace the configured data in the manifest.
 * @param reportUrl
 *           server url.
 * @param appId
 *           application id.
 * @param userData
 *           user custom data (see docs)
 */
+ (BOOL)start:(NSString *)reportUrl
        appId:(NSString *)appId
        error:(NSError **)error;

/**
 * @function start:settings:error
 * @brief Starts the monitoring using a provided start up settings.
 * @param settings
 *        Startup settings, the startup settings can be used along side a settings
 *        file plist or completely replace the file.
 *        If both the startup settings and the plist file contain the key, the 1
 *        from the startup settings will be used.
 */
+ (BOOL)start:(GLAStartupSettings *)settings error:(NSError **)error;

/*!
 * @function stop
 * @brief Stops Glassbox Monitoring.
 */
+ (void)stop;

/*!
 * @function setScreenAsSensitive
 * @brief Exclude certain screen which is associated via controller from Glassbox Monitoring.
 * @param controller
 *              The controller to exclude from Glassbox Monitoring.
 */
+ (void)hideSensitiveScreen:(UIViewController *)screen;

/*!
 * @function setScreenAsSensitive
 * @brief Exclude certain screen which is associated via controller from Glassbox Monitoring.
 * @param controller
 *              The controller to exclude from Glassbox Monitoring.
 */
+ (void)hideSensitiveScreenByName:(NSString *)screenName;

/*!
 * Sets view as sensitive. View info is masked during recording.
 * @param view
 *             the view to mark as Sensitive
 */
+ (void)hideSensitiveView:(UIView *)view;

/*!
 * Sets sensitive view as visible. View info is will not be masked during recording.
 * @param view
 * the view to mark as not Sensitive
 */
+ (void)disableViewHiding:(UIView *)view;

/**
 * Sets view as sensitive with consideration to visibility flags
 * @param view
 * @param ignoreAlphaChannel
 *          indicates that the view should be masked even if it's alpha channel is set to 0
 * @param ignoreHiddenProperty
 *           indicates that the view should be masked even if it's hidden property is set to true
 * @param ignoreBackgroundControllers
 *           indicates that only the top most UIViewController should be masked
 */
+ (void)hideSensitiveView:(UIView *)view visibilityFlags:(GLAVisibilityFlags *)visibilityFlags;

/**
 * @function reportEvent:event:parameters
 *
 * @brief Reports a user defined event.
 *
 * @discussion The user must at least provide an event name.
 *             in case event is nil the method will return and an event won't be reported.
 * @param event
 *          The event name.
 * @param parameters
 *          event parameters, The parameters dictionary can include any of the primitive wrapper classes
 *          or OBJc collections. complex custom object are not supported.
 */
+ (void)reportEvent:(NSString *)event parameters:(NSDictionary *)parameters;

+ (void)reportError:(NSError *)error;

+ (void)setSDKType:(NSString *)stringSdkType;

/**
 * Overrides the Storyboard default screen detection logic by proving a custom screen name.
 * @param screen
 *        custom screen name. nil values are ignored.
 */
+(void)startScreen:(NSString *)screenName;

/**
 * Resets screen name which was provided by the startScreen api.
 * This method is not mandatory, sequential calls to startScreen:screenName does not require a call to endScreen api.
 */
+(void)endScreen;

/**
 *
 Set a delegate to receive callback on session events such as when configuration is received from the server or when it fails
 @param delegate which implements protocol GLASessionDelegate
 */
+ (void)setSessionDelegate:(id <GLASessionDelegate>)delegate;

/**
 *  @function isStarted
 *  @return true if the SDK is currently recording.
 */
+ (BOOL)isStarted;

/**
 *  @function sessionId
 *  @return the SDK session Id (can be used for crashlytics integration for example).
 */
+ (NSString *)sessionId;


/**
 * @function generateSessionLinkWithHost:port:
 * @brief Starts the monitoring with user data.
 * The user data in this case can replace the configured data in the manifest.
 * @param host
 *           Glassbox web interface server url.
 * @param port
 *           Glassbox web interface server port.
*/
+ (NSString *)generateSessionLinkWithHost:(NSString *)host port:(NSInteger)port;

@end


/*!
 * Possible error codes which can be returned (via the NSError) from the start method.
 */
typedef NS_ENUM(NSUInteger, GLAErrorCode) {
    GLASSBOX_UNSUPPORTED_OS_VER = 0,
    GLASSBOX_CONFIG_FILE_MISSING = 1,
    GLASSBOX_SERVER_URL_NOT_CONFIGURED = 2,
    GLASSBOX_APPID_INVALID_FORMAT = 3,
    GLASSBOX_NETWORK_ERROR = 4,
    GLASSBOX_INVALID_JSON_FORMAT = 5,
    GLASSBOX_SETTINGS_ERROR = 6
};
